/* Copyright (C) 2004 David Decotigny

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA. 
*/
#ifndef _SOS_KTHREAD_H_
#define _SOS_KTHREAD_H_

#include <sos/errno.h>

/**
 * @file kthread.h
 *
 * SOS Kernel thread management API
 */


/* Forward declaration */
struct sos_kthread;

#include <hwcore/cpu_context.h>
#include <sos/sched.h>
#include <sos/kwaitq.h>
#include <sos/time.h>


/**
 * The possible states of a valid kernel thread
 */
typedef enum { SOS_KTHR_CREATED, /**< Thread created, not fully initialized */
	       SOS_KTHR_READY,   /**< Thread fully initialized or waiting
                                      for CPU after having been blocked */
	       SOS_KTHR_RUNNING, /**< Thread currently running on CPU */
	       SOS_KTHR_BLOCKED, /**< Thread waiting for I/O (+ in at LEAST
				      one kwaitq) and/or sleeping (+ in NO
				      kwaitq) */
	       SOS_KTHR_ZOMBIE,  /**< Thread terminated execution, waiting to
				      be deleted by kernel */
             } sos_kthread_state_t;


/**
 * TCB (Thread Control Block): structure describing a Kernel
 * thread. Don't access these fields directly: prefer using the
 * accessor functions below.
 */
struct sos_kthread
{
#define SOS_KTHR_MAX_NAMELEN 32
  char name[SOS_KTHR_MAX_NAMELEN];

  sos_kthread_state_t  state;
  sos_sched_priority_t priority;

  /* The hardware context of the thread */
  struct sos_cpu_kstate *cpu_kstate;
  sos_vaddr_t stack_base_addr;
  sos_size_t  stack_size;

  /* Data specific to each state */
  union
  {
    struct
    {
      struct sos_sched_queue *rdy_queue;
      struct sos_kthread     *rdy_prev, *rdy_next;
    } ready;
  }; /* Anonymous union (gcc extenion) */


  /*
   * Data used by the kwaitq subsystem: list of kwaitqueues the thread
   * is waiting for.
   *
   * @note: a RUNNING or READY thread might be in one or more
   * waitqueues ! The only property we have is that, among these
   * waitqueues (if any), _at least_ one has woken the thread.
   */
  struct sos_kwaitq_entry *kwaitq_list;


  /**
   * Chaining pointers for global ("gbl") list of threads (debug)
   */
  struct sos_kthread *gbl_prev, *gbl_next;
};


/**
 * Definition of the function executed by a kernel thread
 */
typedef void (*sos_kthread_start_routine_t)(void *arg);


/**
 * Initialize the subsystem responsible for kernel thread management
 *
 * Initialize primary kernel thread so that it can be handled the same
 * way as an ordinary thread created by sos_kthread_create().
 */
sos_ret_t sos_kthread_subsystem_setup(sos_vaddr_t init_thread_stack_base_addr,
				      sos_size_t init_thread_stack_size);


/**
 * Create a new kernel thread
 */
struct sos_kthread *sos_kthread_create(const char *name,
				       sos_kthread_start_routine_t start_func,
				       void *start_arg,
				       sos_sched_priority_t priority);


/**
 * Terminate the execution of the current thread. Called by default
 * when the start routine returns.
 */
void sos_kthread_exit() __attribute__((noreturn));


/**
 * Get the identifier of the thread currently running on CPU. Trivial
 * function.
 */
struct sos_kthread *sos_kthread_get_current();


/**
 * If thr == NULL, set the priority of the current thread. Trivial
 * function.
 *
 * @note NOT protected against interrupts
 */
sos_sched_priority_t sos_kthread_get_priority(struct sos_kthread *thr);


/**
 * If thr == NULL, get the state of the current thread. Trivial
 * function.
 *
 * @note NOT protected against interrupts
 */
sos_kthread_state_t sos_kthread_get_state(struct sos_kthread *thr);


/**
 * If thr == NULL, set the priority of the current thread
 *
 * @note NO context-switch ever occurs in this function !
 */
sos_ret_t sos_kthread_set_priority(struct sos_kthread *thr,
				   sos_sched_priority_t priority);


/**
 * Yield CPU to another ready thread.
 *
 * @note This is a BLOCKING FUNCTION
 */
sos_ret_t sos_kthread_yield();


/**
 * Release the CPU for (at least) the given delay.
 *
 * @param delay The delay to wait for. If delay == NULL then wait
 * forever that any event occurs.
 *
 * @return SOS_OK when delay expired (and delay is reset to zero),
 * -SOS_EINTR otherwise (and delay contains the amount of time
 * remaining).
 *
 * @note This is a BLOCKING FUNCTION
 */
sos_ret_t sos_kthread_sleep(/* in/out */struct sos_time *delay);


/**
 * Mark the given thread as READY (if not already ready) even if it is
 * blocked in a kwaitq or in a sleep ! As a result, the interrupted
 * kwaitq/sleep function call of the thread will return with
 * -SOS_EINTR.
 *
 * @return -SOS_EINVAL if thread does not exist, or -SOS_EFATAL if
 * marked ZOMBIE.
 *
 * @note As a result, the semaphore/mutex/conditions/... functions
 * return values SHOULD ALWAYS be checked ! If they are != SOS_OK,
 * then the caller should consider that the resource is not aquired
 * because somebody woke the thread by some way.
 */
sos_ret_t sos_kthread_force_unblock(struct sos_kthread *kthread);


#endif /* _SOS_KTHREAD_H_ */
